/**
 * @file Social.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials;

import com.avaje.ebean.validation.Length;
import com.avaje.ebean.validation.NotEmpty;

import java.io.Serializable;
import java.util.Date;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

import org.bukkit.ChatColor;
import org.bukkit.entity.Player;
import org.mudcraft.bukkit.socials.message.Message;

/**
 * Represents a pre-configured emote a player can perform.
 * @author Geoffrey Davis
 */
@Entity
@Table(name="social")
public class Social implements Cloneable, Serializable {
	/**
	 * Constructs a new {@link Social} instance.
	 */
	public Social() {
		// Nothing.
	}
	
	/**
	 * Constructs a new {@link Social} instance.
	 * @param other the {@link Social} whose state is to be copied
	 */
	public Social(Social other) {
	    if (other != null) {
	        this.actorMessage = other.actorMessage;
	        this.created = cloneDate(other.created);
	        this.createdBy = other.createdBy;
	        this.id = other.id;
	        this.modified = cloneDate(other.modified);
	        this.modifiedBy = other.modifiedBy;
	        this.noArgumentMessage = other.noArgumentMessage;
	        this.name = other.name;
	        this.targetMessage = other.targetMessage;
	    }
	}

	/**
	 * Creates a copy of this object.
	 * @return a copy of this object
	 * @see Cloneable
	 * @see Object#clone()
	 */
	public Object clone() {
        // Return a copy of this object.
	    return new Social(this);
	}
	
    /**
     * Creates a copy of a {@link Date}.
     * @param date the {@link Date} to be cloned
     * @return a copy of the specified date, or <code>null</code>
     */
    private final static Date cloneDate(Date date) {
        return date != null ? new Date(date.getTime()) : null;
    }
	
    /**
     * Gets whether this object is equal to another object.
     * @param o the object to which this object is to be compared
     * @return <code>true</code> if the specified object is equivalent
     * @see Object#equals(Object)
     */
    public boolean equals(Object o) {
        if (o instanceof Social) {
            // Cast the object to a Social.
            final Social oSocial = (Social) o;
            
            if (name == oSocial.name) {
                return true;
            } else if (name != null) {
                // Compare our name to the other social's.
                if (name.equalsIgnoreCase(oSocial.name)) {
                    return true;
                }
            } else if (oSocial.name != null) {
                // Compare the other social's name to ours.
                if (oSocial.name.equalsIgnoreCase(name)) {
                    return true;
                }
            }
        }
        return false;
    }
    
	/**
	 * Gets the message shown when the actor performing the social is
	 * also the target of the social.
	 * @return the social's actor message
	 * @see #setActorMessage(String)
	 */
	public String getActorMessage() {
		return actorMessage;
	}
	
   /**
     * Gets the creation time stamp.
     * @return the creation time stamp
     * @see #setCreated(Date)
     */
    public Date getCreated() {
        return cloneDate(created);
    }
    
    /**
     * Gets the name of the player who created the social.
     * @return the name of the player who created the social
     * @see #setCreatedBy(String)
     */
    public String getCreatedBy() {
        return createdBy;
    }
    
    /**
     * Gets the record identity.
     * @return the record identity
     * @see #setId(Integer)
     */
    public Integer getId() {
        return id;
    }
    
	/**
     * Gets the modification time stamp.
     * @return the modification time stamp
     * @see #setModified(Date)
     */
    public Date getModified() {
        return cloneDate(modified);
    }
    
    /**
     * Gets the name of the player who modified the social.
     * @return the name of the player who modified the social
     * @see #setModifiedBy(String)
     */
    public String getModifiedBy() {
        return modifiedBy;
    }
    
	/**
	 * Gets the social's name.
	 * @return the social's name
	 * @see #setName(String)
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Gets the message shown when the actor does not specify a target.
	 * @return the social's no argument message
	 * @see #setNoArgumentMessage(String)
	 */
	public String getNoArgumentMessage() {
		return noArgumentMessage;
	}
	
	/**
	 * Gets the message shown when the actor specifies a target.
	 * @return the social's target message
	 * @see #setTargetMessage(String)
	 */
	public String getTargetMessage() {
		return targetMessage;
	}
	
	/**
	 * Calculates a hashcode for this object.
	 * @return a hashcode for this object
	 * @see Object#hashCode()
	 */
	public int hashCode() {
		// Begin with a prime number.
		int hashCode = 97;
		
		// Multiply by a prime number and add `name` hashcode.
		hashCode = (hashCode * 101) +
			(name != null ? name.toLowerCase().hashCode() : 0);
			
		// Return the calculated hashcode.
		return hashCode;
	}
	
	/**
	 * Joins an array of strings.
	 * @param split the array of {@link String}s to be joined
	 * @return a {@link String} composed of the elements of the specified
	 *     array separated by spaces 
	 */
	private final static String join(String[] split) {
	    // Create a StringBuilder instance.
	    final StringBuilder stringBuilder = new StringBuilder();
	    
	    // Iterate over the elements of the array.
	    for (String string: split) {
	        if (string.length() != 0) {
    	        if (stringBuilder.length() != 0) {
    	            // Add spaces between the elements.
    	            stringBuilder.append(' ');
    	        }
    	        // Append the next element.
    	        stringBuilder.append(string);
	        }
	    }
	    return stringBuilder.toString();
	}
    
    /**
     * Causes the {@link Player} specified by <code>actor</code> to perform
     * this social.
     * @param plugin the {@link Socials} object
     * @param actor the {@link Player} performing the social
     * @param split an array containing any arguments specified by the
     *      player at the name line
     */
    public void perform(Socials plugin, Player actor, String[] split) {
        // Join the arguments to form a single argument.
        final String argument = join(split);
        
        // The player who is the social's target.
        Player target = null;
        
        // The words "me", "myself", and "self" indicate the player who
        // is performing the social.
        if (argument.equalsIgnoreCase("me") ||
            argument.equalsIgnoreCase("myself") ||
            argument.equalsIgnoreCase("self")) {
            target = actor;
        } else if (argument.length() != 0){
            // Attempt to match the player whose name was specified.
            final List<Player> matches = actor.getServer().matchPlayer(
                    argument
                    );
            
            if (matches.size() != 0) {
                // Always select the first match.
                target = matches.get(0);
            }
        }
        // Create a Message object.
        final Message message = new Message(plugin);
        
        // Populate the message context.
        message.getContext().setString("line", argument);
        message.getContext().setPlayer("me", actor);
        message.getContext().setString("social", name);
        message.getContext().setPlayer("target", target);
        
        // Set the default chat color to light purple.
        message.setChatColor(ChatColor.LIGHT_PURPLE);
        
        if (target != null && target != actor) {
            // Make certain the social has a target message.
            if (targetMessage != null && targetMessage.length() != 0) {
                // Configure the message's format specifier.
                message.setFormatSpecifier(targetMessage);
            } else {
                // Tell the player they can't specify a target.
                actor.sendMessage(ChatColor.RED + "You can't specify a target.");
            }
        } else if (target != null && target == actor) {
            // Make certain the social has an actor message.
            if (actorMessage != null && actorMessage.length() != 0) {
                // Configure the message's format specifier.
                message.setFormatSpecifier(actorMessage);
            } else {
                // Tell the player they can't specify a target.
                actor.sendMessage(ChatColor.RED + "You can't specify yourself.");
            }
        } else {
            // Make certain the social has a zero-argument message.
            if (noArgumentMessage != null && noArgumentMessage.length() != 0) {
                // Configure the message's format specifier.
                message.setFormatSpecifier(noArgumentMessage);
            } else {
                // Tell the player they must specify a target.
                actor.sendMessage(ChatColor.RED + "You must specify a target.");
            }
        }
        // If the format specifier was set we're good to go.
        if (message.getFormatSpecifier() != null &&
            message.getFormatSpecifier().length() != 0) {
            // Dispatch the message to players.
            message.send(actor);
        }
    }
    
	/**
	 * Sets the message shown when the actor performing the social
	 * is also the target of the social.
	 * @param actorMessage the social's actor message
	 * @see #getActorMessage()
	 */
	public void setActorMessage(String actorMessage) {
	    this.actorMessage = null;
	    if (actorMessage != null && actorMessage.trim().length() != 0) {
	        this.actorMessage = actorMessage.trim();
	    }
	}
	
    /**
     * Sets the creation time stamp.
     * @param created the creation time stamp
     * @see #getCreated()
     */
    public void setCreated(Date created) {
        this.created = cloneDate(created);
    }
    
    /**
     * Sets the name of the player that created the social.
     * @param createdBy the name of the player that created the social
     * @see #getCreatedBy()
     */
    public void setCreatedBy(String createdBy) {
        this.createdBy = null;
        if (createdBy != null && createdBy.trim().length() != 0) {
            this.createdBy = createdBy.trim().toLowerCase();
        }
    }
    
    /**
     * Sets the record identity.
     * @param id the record identity
     * @see #getId()
     */
    public void setId(Integer id) {
        this.id = id;
    }
    
    /**
     * Sets the modification time stamp.
     * @param modified the modification time stamp
     * @see #getModified()
     */
    public void setModified(Date modified) {
        this.modified = cloneDate(modified);
    }
	
    /**
     * Sets the name of the player that modified the social.
     * @param modifiedBy the name of the player that modified the social
     * @see #getModifiedBy()
     */
    public void setModifiedBy(String modifiedBy) {
        this.modifiedBy = null;
        if (modifiedBy != null && modifiedBy.trim().length() != 0) {
            this.modifiedBy = modifiedBy.trim().toLowerCase();    
        }
    }
    
	/**
	 * Sets the social's name.
	 * @param name the social's name
	 * @see #getName()
	 */
	public void setName(String name) {
	    this.name = null;
	    if (name != null && name.trim().length() != 0) {
	        this.name = name.trim().toLowerCase();
	    }
	}
	
	/**
	 * Sets the message shown when the actor does not specify a target.
	 * @param noArgumentMessage the social's no argument message
	 * @see #getNoArgumentMessage()
	 */
	public void setNoArgumentMessage(String noArgumentMessage) {
	    this.noArgumentMessage = null;
	    if (noArgumentMessage != null &&
	        noArgumentMessage.trim().length() != 0) {
	        this.noArgumentMessage = noArgumentMessage.trim();    
	    }
	}
	
	/**
	 * Sets the message shown when the actor specifies a target.
	 * @param targetMessage the social's target message
	 * @see #getTargetMessage()
	 */
	public void setTargetMessage(String targetMessage) {
	    this.targetMessage = null;
	    if (targetMessage != null && targetMessage.trim().length() != 0) {
	        this.targetMessage = targetMessage.trim();
	    }
	}
	
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object
     * @see Object#toString()
     */
    public String toString() {
        // Create a string builder. 
        final StringBuilder stringBuilder = new StringBuilder();
        
        // Write the contents of the social to the string builder.
        stringBuilder.
            append(getClass().getSimpleName()).append('{').
            append("actorMessage=").append(actorMessage).append(", ").
            append("created=").append(created).append(", ").
            append("createdBy=").append(createdBy).append(", ").
            append("id=").append(id).append(", ").
            append("modified=").append(modified).append(", ").
            append("modifiedBy=").append(modifiedBy).append(", ").
            append("name=").append(name).append(", ").
            append("noArgumentMessage=").append(noArgumentMessage).append(", ").
            append("targetMessage=").append(targetMessage).
            append('}');

        // Return the contents of the string builder.
        return stringBuilder.toString();
    }
    
	/**
	 * The message shown when the actor performing the social is also
	 * the target of the social.
	 * @see #getActorMessage()
	 * @see #setActorMessage(String)
	 */
	@Length(max=255)
	private String actorMessage;
	
	/**
	 * The creation time stamp.
	 * @see #getCreated()
	 * @see #setCreated(Date)
	 */
	@Temporal(TemporalType.TIMESTAMP)
	private Date created;

	/**
     * The name of player who created the social.
     * @see #getCreatedBy()
     * @see #setCreatedBy(String)
     */
    @Length(max=255)
	private String createdBy;
	
	/**
	 * The record identity.
	 * @see #getId()
	 * @see #setId(Integer)
	 */
	@Id
	private Integer id;
	
	/**
     * The modification time stamp.
     * @see #getModified()
     * @see #setModified(Date)
     */
	@Temporal(TemporalType.TIMESTAMP)
    private Date modified;

    /**
     * The name of the player who modified the social.
     * @see #getModifiedBy()
     * @see #setModifiedBy(String)
     */
    @Length(max=255)
    private String modifiedBy;
    
	/**
	 * The social's name.
	 * @see #getName()
	 * @see #setName(String)
	 */
    @Column(unique=true)
    @Length(max=255)
    @NotEmpty
	private String name;
	
	/**
	 * The message shown when the actor does not specify a target.
	 * @see #getNoArgumentMessage()
	 * @see #setNoArgumentMessage(String)
	 */
    @Length(max=255)
	private String noArgumentMessage;
	
	/**
	 * A unique version serial number for object serialization.
	 * @see {@link Serializable}
	 */
	private static final long serialVersionUID = 2602769714157893737L;
	
	/**
	 * The message shown when the actor specifies a target.
	 * @see #getTargetMessage()
	 * @see #setTargetMessage(String)
	 */
    @Length(max=255)
	private String targetMessage;
}
