/**
 * @file ProfileCommand.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials.commands;

import java.text.DateFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.bukkit.ChatColor;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.mudcraft.bukkit.socials.Socials;
import org.mudcraft.bukkit.socials.user.Gender;
import org.mudcraft.bukkit.socials.user.SocialUser;

/**
 * The /profile command.
 * @author Geoffrey Davis
 */
public class ProfileCommand implements CommandExecutor {
   /**
     * Constructs a new {@link ProfileCommand} instance.
     * @param plugin the {@link Socials} object
     */
    public ProfileCommand(Socials plugin) {
        if (plugin == null) {
            // Throw an exception to indicate an error.
            throw new IllegalArgumentException("plugin is null");
        }
        this.plugin = plugin;
    }
    
    /**
     * Gets the {@link Socials} object.
     * @return the {@link Socials} object
     */
    public Socials getPlugin() {
        return plugin;
    }
    
    /**
     * Joins a collection of strings.
     * @param collection the {@link String}s to be joined together
     * @return the joined {@link String}
     */
    private static String join(Iterable<String> collection) {
        // Create a string builder to contain the joined string.
        final StringBuilder stringBuilder = new StringBuilder();
        
        // Iterate over the contents of the collection.
        for (String string: collection) {
            // Skip empty strings.
            if (string != null && string.length() != 0) {
                if (stringBuilder.length() != 0) {
                    // Insert a space between each word.
                    stringBuilder.append(' ');
                }
                // Append the next word to the string.
                stringBuilder.append(string);
            }
        }
        return stringBuilder.toString();
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean onCommand(
            CommandSender sender, Command command, String label,
            String[] split) {
        // Create a queue of the command line parameters.
        final Queue<String> splitQueue = new LinkedList<String>(
                Arrays.asList(split)
                );

        // Some information we want to keep track of.
        String fieldName = null;
        Player target = null;
        String targetName = null;

        // Check the length of the queue.
        if (splitQueue.size() != 0) {
            // Read one word from the queue.
            final String oneWord = splitQueue.remove().trim();

            // Determine whether a player name was specified.
            if (oneWord.charAt(0) == '@') {
                // The target's name follows the at sign.
                targetName = oneWord.substring(1);
                
                // Check the length of the queue again.
                if (splitQueue.size() != 0) {
                    // Remove the next element from the queue.
                    fieldName = splitQueue.remove();
                }
            } else {
                // The next word is the field name.
                fieldName = oneWord;
            }
        }
        // Determine if a target's name was specified.
        if (targetName != null && targetName.length() != 0) {
            // Search for the player by name.
            final List<Player> matches =
                    getPlugin().getServer().matchPlayer(targetName);
            
            if (matches.size() != 0) {
                // Get the first match we found.
                target = matches.get(0);
            } else {
                // Tell the command sender the player doesn't exist.
                sender.sendMessage(ChatColor.RED + "No such player!");
                return true;
            }
        // Make certain the command sender is a player before we use it.
        } else if (sender instanceof Player) {
            target = (Player) sender;
        }
        if (target == null) {
            // Tell the command sender they have to specify someone.
            sender.sendMessage(ChatColor.RED + "You must specify a player!");
            return true;
        }
        // We'll set a field if a field name was specified.
        if (fieldName != null && fieldName.length() != 0) {
            try {
                // If this succeeds it's a /profile [page] command.
                splitQueue.add(Integer.toString(Integer.parseInt(fieldName)));
            } catch (NumberFormatException ex) {
                // Check for the "help" sub-command.
                if (fieldName.equalsIgnoreCase("help")) {
                    // Display some help information.
                    return profileHelp(sender, join(splitQueue));
                } else {
                    // Set the corresponding field.
                    return profileField(
                            sender, target, fieldName, join(splitQueue));
                }
            }
        }
        // Otherwise display the profile of the specified target.
        return profileDisplay(sender, target, join(splitQueue));
    }
    
    /**
     * Displays a player's profile.
     * @param sender the {@link CommandSender} object
     * @param target the {@link Player} whose profile is to be displayed
     * @return <code>true</code> if the operation was successful
     */
    protected boolean profileDisplay(
                CommandSender sender, Player target, String argument) {
        // Determine whether privacy is enforced.
        final boolean privacyCheck = sender.isOp() || sender.equals(target);
        
        // Get the social user for the specified player.
        final SocialUser user = getPlugin().loadOrCreateSocialUser(target);
        
        // Create a list to contain the profile elements.
        final List<String> profileElements = new ArrayList<String>();
        
        // Begin with page #1.
        int pageNumber = 1;
        
        // Determine if an argument was specified.
        if (argument != null && argument.trim().length() != 0) {
            try {
                // Parse the page number.
                pageNumber = Integer.parseInt(argument);
            } catch (NumberFormatException ex) {
                // Tell the sender they entered an illegal page number.
                sender.sendMessage(ChatColor.RED + "Illegal page number.");
                return true;
            }
        }

        // Add the player's display name.
        profileElements.add(ChatColor.DARK_GREEN + "Display Name: " + ChatColor.GREEN + target.getDisplayName());

        // Add the player's entity ID number.
        profileElements.add(ChatColor.DARK_GREEN + "Entity ID Number: " + ChatColor.GREEN + target.getEntityId());
        
        // Add the player's `gender` field.
        profileElements.add(ChatColor.DARK_GREEN + "Gender: " + ChatColor.GREEN + user.getGender().toString().toLowerCase());
        
        // Check privacy and profile options.
        if (privacyCheck || user.getShowRealName()) {
            // Add the player's `realName` field.
            profileElements.add(ChatColor.DARK_GREEN + "Real Name: " + ChatColor.GREEN + user.getRealName());
        }
        // Check privacy and profile options.
        if (privacyCheck || user.getShowRealGender()) {
            // Add the player's `realGender` field.
            profileElements.add(ChatColor.DARK_GREEN + "Real Gender: " + ChatColor.GREEN + user.getRealGender());
        }
        // Check privacy and profile options.
        if (privacyCheck || user.getShowEmailAddress()) {
            // Add the player's `emailAddress` field.
            profileElements.add(ChatColor.DARK_GREEN + "Email Address: " + ChatColor.GREEN + user.getEmailAddress());
        }
        // Check privacy and profile options.
        if (user.getBirthDate() == null || !user.getShowBirthday()) {
            // Add the player's `birthDate` field.
            profileElements.add(ChatColor.DARK_GREEN + "Birthday: " + ChatColor.GREEN + "Not Disclosed");
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showAge` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Age: " + ChatColor.GREEN + user.getShowAge());
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showBirthday` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Birthday: " + ChatColor.GREEN + user.getShowBirthday());
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showBirthYear` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Birth Year: " + ChatColor.GREEN + user.getShowBirthYear());
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showEmailAddress` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Email Address: " + ChatColor.GREEN + user.getShowEmailAddress());
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showRealGender` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Real Gender: " + ChatColor.GREEN + user.getShowRealGender());
        }
        // Check privacy only.
        if (sender.equals(target)) {
            // Add the player's `showRealName` field.
            profileElements.add(ChatColor.DARK_GREEN + "Show Real Name: " + ChatColor.GREEN + user.getShowRealName());
        }
        // Calculate the maximum number of pages.
        final int pageCount = profileElements.size() / 9 + 1;
        
        // Force the selected page to be in range.
        pageNumber = Math.max(Math.min(pageNumber, pageCount), 1);
        
        // Send the social profile header.
        sender.sendMessage(ChatColor.DARK_GREEN + "Social Profile for " +
                           ChatColor.GREEN + target.getName() +
                           ChatColor.DARK_GREEN + " (Page " +
                           ChatColor.GREEN + pageNumber +
                           ChatColor.DARK_GREEN + " of " +
                           ChatColor.GREEN + pageCount +
                           ChatColor.DARK_GREEN + ")");
        
        // Determine the begin and end indexes in the list.
        final int beginIndex = (pageNumber - 1) * 9;
        final int endIndex = Math.min(profileElements.size(), beginIndex + 9);
        
        // Iterate over the selected profile elements.
        for (int i = beginIndex; i < endIndex; ++i) {
            // Send the corresponding profile elements to the sender.
            sender.sendMessage(profileElements.get(i));
        }
        return true;
    }
    
    /**
     * Displays help information.
     * @param sender the {@link CommandSender} object
     * @param argument the remaining arguments passed on the chat line
     * @return <code>true</code> if the operation was successful
     */
    protected boolean profileHelp(CommandSender sender, String argument) {
        // Create a list to contain the help elements.
        final List<String> helpElements = new ArrayList<String>();
        
        // Begin with page #1.
        int pageNumber = 1;
        
        // Determine if an argument was specified.
        if (argument != null && argument.trim().length() != 0) {
            try {
                // Parse the page number.
                pageNumber = Integer.parseInt(argument);
            } catch (NumberFormatException ex) {
                // Tell the sender they entered an illegal page number.
                sender.sendMessage(ChatColor.RED + "Illegal page number.");
                return true;
            }
        }
        
        // Add the help elements to the list.
        helpElements.add(" - birthDate [value] - Sets player's birth date.");
        helpElements.add(" - emailAddress [value] - Sets player's email address.");
        helpElements.add(" - gender [female/male] - Sets player's gender.");
        helpElements.add(" - realGender [female/male] - Sets player's actual gender.");
        helpElements.add(" - realName [value] - Sets player's actual name.");
        helpElements.add(" - showAge [true/false] - Sets age flag.");
        helpElements.add(" - showBirthDate [true/false] - Sets birth-date flag.");
        helpElements.add(" - showBirthYear [true/false] - Sets birth-year flag.");
        helpElements.add(" - showEmailAddress [true/false] - Sets email-address flag.");
        helpElements.add(" - showRealGender [true/false] - Sets real-gender flag.");
        helpElements.add(" - showRealName [true/false] - Sets real-name flag.");
        
        // Calculate the maximum number of pages.
        final int pageCount = helpElements.size() / 8 + 1;
        
        // Force the selected page to be in range.
        pageNumber = Math.max(Math.min(pageNumber, pageCount), 1);
        
        sender.sendMessage(ChatColor.DARK_GREEN + "Profile Help Page (page " + ChatColor.GREEN + pageNumber + ChatColor.DARK_GREEN + " of " + ChatColor.GREEN + pageCount + ChatColor.DARK_GREEN + ")");
        sender.sendMessage(ChatColor.DARK_GREEN + "Describes profile fields and options.");
        
        // Determine the begin and end indexes in the list.
        final int beginIndex = (pageNumber - 1) * 8;
        final int endIndex = Math.min(helpElements.size(), beginIndex + 8);
        
        // Iterate over the selected help elements.
        for (int i = beginIndex; i < endIndex; ++i) {
            // Send the corresponding help elements to the sender.
            sender.sendMessage(ChatColor.GRAY + helpElements.get(i));
        }
        return true;
    }
    
    /**
     * Sets a field of the target {@link Player}'s profile.
     * @param sender the {@link CommandSender} object
     * @param target the {@link Player} whose profile is to have a field set
     * @param fieldName the name of the field to be set
     * @param fieldValue the value to be assigned to the specified field
     * @return <code>true</code> if the operation was successful
     */
    protected boolean profileField(
            CommandSender sender, Player target, String fieldName,
            String fieldValue) {
        // Load the social user for the target player.
        final SocialUser user = getPlugin().loadOrCreateSocialUser(target);
        
        // Make certain we're allowed to set the field.
        if (!sender.equals(target) && !sender.isOp()) {
            // Tell the command sender they don't have permission.
            sender.sendMessage(ChatColor.RED + "You aren't allowed to set that!");
        // The `birthDate` field.
        } else {
            if (fieldName.equalsIgnoreCase("birthDate")) {
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    try {
                        // Get the date format object.
                        final DateFormat dateFormat =
                            DateFormat.getDateInstance(DateFormat.SHORT);
                        // Attempt to parse the field value into a date.
                        user.setBirthDate(dateFormat.parse(fieldValue.trim()));
                    } catch (ParseException ex) {
                        sender.sendMessage(ChatColor.RED + "Illegal date syntax.");
                    }
                } else {
                    user.setBirthDate(null);
                }
             // The `emailAddress` field.
            } else if (fieldName.equalsIgnoreCase("emailAddress")) {
                user.setEmailAddress(null);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    user.setEmailAddress(fieldValue.trim());
                }
            // The `gender` field.
            } else if (fieldName.equalsIgnoreCase("gender")) {
                user.setGender(Gender.NONE);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    if (fieldValue.equalsIgnoreCase("f")) {
                        user.setGender(Gender.FEMALE);
                    } else if (fieldValue.equalsIgnoreCase("female")) {
                        user.setGender(Gender.FEMALE);
                    } else if (fieldValue.equalsIgnoreCase("m")) {
                        user.setGender(Gender.MALE);
                    } else if (fieldValue.equalsIgnoreCase("male")) {
                        user.setGender(Gender.MALE);
                    }
                }
            // The `realGender` field.
            } else if (fieldName.equalsIgnoreCase("realGender")) {
                user.setRealGender(Gender.NONE);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    if (fieldValue.equalsIgnoreCase("f")) {
                        user.setRealGender(Gender.FEMALE);
                    } else if (fieldValue.equalsIgnoreCase("female")) {
                        user.setRealGender(Gender.FEMALE);
                    } else if (fieldValue.equalsIgnoreCase("m")) {
                        user.setRealGender(Gender.MALE);
                    } else if (fieldValue.equalsIgnoreCase("male")) {
                        user.setRealGender(Gender.MALE);
                    }
                }
            // The `realName` field.
            } else if (fieldName.equalsIgnoreCase("realName")) {
                user.setRealName(null);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    user.setRealName(fieldValue.trim());
                }
            // The `showAge` field.
            } else if (fieldName.equalsIgnoreCase("showAge")) {
                user.setShowAge(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowAge(booleanValue);
                }
            // The `showBirthday` field.
            } else if (fieldName.equalsIgnoreCase("showBirthday")) {
                user.setShowBirthday(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowBirthday(booleanValue);
                }
            // The `showBirthYear` field.
            } else if (fieldName.equalsIgnoreCase("showBirthYear")) {
                user.setShowBirthYear(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowBirthYear(booleanValue);
                }
            // The `showEmailAddress` field.
            } else if (fieldName.equalsIgnoreCase("showEmailAddress")) {
                user.setShowEmailAddress(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowEmailAddress(booleanValue);
                }
            // The `showRealGender` field.
            } else if (fieldName.equalsIgnoreCase("showRealGender")) {
                user.setShowRealGender(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowRealGender(booleanValue);
                }
            // The `showRealName` field.
            } else if (fieldName.equalsIgnoreCase("showRealName")) {
                user.setShowRealName(false);
                if (fieldValue != null && fieldValue.trim().length() != 0) {
                    final boolean booleanValue = Boolean.parseBoolean(
                            fieldValue.trim()
                            );
                    user.setShowRealName(booleanValue);
                }
            } else {
                return false;
            }
            // Save changes to the social user's properties.
            getPlugin().saveSocialUser(user);
        }
        return true;
    }

    
    /**
     * The {@link Socials} object.
     * @see {@literal #getPlugin()}
     */
    private final Socials plugin;
}
