/**
 * @file SocialCommand.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials.commands;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import org.bukkit.ChatColor;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.mudcraft.bukkit.socials.Social;
import org.mudcraft.bukkit.socials.Socials;
import org.mudcraft.bukkit.socials.message.WordWrap;

/**
 * The /social command.
 * @author Geoffrey Davis
 */
public class SocialCommand implements CommandExecutor {
    /**
     * Constructs a new {@link SocialCommand} instance.
     * @param plugin the {@link Socials} object
     */
    public SocialCommand(Socials plugin) {
        if (plugin == null) {
            // Throw an exception to indicate an error.
            throw new IllegalArgumentException("plugin is null");
        }
        this.plugin = plugin;
    }
    
    /**
     * Gets the {@link Socials} object.
     * @return the {@link Socials} object
     */
    public Socials getPlugin() {
        return plugin;
    }
    
    /**
     * Returns a textual representation of a {@link Date}.
     * @param date the {@link Date} object
     * @return the textual representation of the specified {@link Date},
     *      or the word "None"
     */
    protected String dateString(Date date) {
        if (date != null) {
            // Use a fairly simple format to format the date.
            return DateFormat.getDateTimeInstance(
                    DateFormat.MEDIUM,
                    DateFormat.SHORT).
                    format(date);
        }
        return "None";
    }
    
    /**
     * Lists one social.
     * @param sender the {@link CommandSender} object
     * @param socialName the {@link Social}'s canonical name 
     * @return <code>true</code> if the operation was successful
     */
    protected boolean listOneSocial(
            CommandSender sender, String socialName) {
        // Search for the social by name.
        final Social social = getPlugin().loadSocial(socialName);

        if (social == null) {
            sender.sendMessage(ChatColor.RED + "No such social!");
        } else {
            sender.sendMessage(ChatColor.DARK_GREEN + "Social: " + ChatColor.GREEN + social.getName() + ChatColor.DARK_GREEN + " (#" + ChatColor.GREEN + social.getId() + ChatColor.DARK_GREEN + ")");
            sender.sendMessage(ChatColor.DARK_GREEN + "Created: " + ChatColor.GREEN + dateString(social.getCreated()));
            sender.sendMessage(ChatColor.DARK_GREEN + "Created By: " + ChatColor.GREEN + stringOrNone(social.getCreatedBy()));
            sender.sendMessage(ChatColor.DARK_GREEN + "Modified: " + ChatColor.GREEN + dateString(social.getModified()));
            sender.sendMessage(ChatColor.DARK_GREEN + "Modified By: " + ChatColor.GREEN + stringOrNone(social.getModifiedBy()));
            sender.sendMessage(ChatColor.DARK_GREEN + "Actor Message: " + ChatColor.GREEN + stringOrNone(social.getActorMessage()));
            sender.sendMessage(ChatColor.DARK_GREEN + "No-Argument Message: " + ChatColor.GREEN + stringOrNone(social.getNoArgumentMessage()));
            sender.sendMessage(ChatColor.DARK_GREEN + "Target Message: " + ChatColor.GREEN + stringOrNone(social.getTargetMessage()));
        }
        return true;
    }
    
    /**
     * Lists out the configured socials.
     * @param sender the {@link CommandSender} object
     * @return <code>true</code> if the operation was successful
     */
    protected boolean listSocials(CommandSender sender) {
        // Search for any defined socials.
        final List<Social> resultSet = getPlugin().getDatabase().
                find(Social.class).
                orderBy("name").
                findList();
        
        if (resultSet.size() == 0) {
            // Tell the command sender there are no socials.
            sender.sendMessage(ChatColor.RED + "No socials are defined.");
        } else {
            // Create a string builder.
            final StringBuilder stringBuilder = new StringBuilder();
            
            // Iterate over the results of the query.
            for (Social result: resultSet) {
                if (stringBuilder.length() != 0) {
                    // Append a comma between each social.
                    stringBuilder.
                        append(ChatColor.DARK_GREEN).append(',').append(' ');
                }
                // Append the name of the social.
                stringBuilder.append(ChatColor.GREEN).append(result.getName());
            }
            
            // Begin with some introductory text.
            stringBuilder.
                insert(0, "The following socials are available: ").
                insert(0, ChatColor.DARK_GREEN);

            // Wrap the contents of the string builder.
            String[] lines = WordWrap.wrap(stringBuilder.toString());
            
            // Iterate over the lines that were returned.
            for (String line: lines) {
                // Send the line to the command sender.
                sender.sendMessage(line);
            }
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean onCommand(
            CommandSender sender, Command command, String label,
            String[] split) {
        boolean retcode = false;
        if (split.length != 0) {
            // Obtain a list for the split array.
            final List<String> splitList = new ArrayList<String>(
                    Arrays.asList(split)
                    );

            // Record the name of the sub-command.
            final String subcommand = splitList.remove(0);

            // Check the 'list' sub-command.
            if (subcommand.equalsIgnoreCase("list")) {
                if (splitList.size() == 0) {
                    // List the socials to the command sender.
                    retcode = listSocials(sender);
                } else {
                    // Determine the name of the social.
                    final String socialName = splitList.remove(0);
                    
                    if (splitList.size() == 0) {
                        // Display the social to the command sender.
                        retcode = listOneSocial(sender, socialName);
                    }
                }
            }
        }
        return retcode;
    }
    
    /**
     * Returns the specified string or the word "None" if the string has
     * a value of <code>null</code> or a (trimmed) length of zero.
     * @param string the string to be evaluated
     * @return the specified string or the word "None" if the string has
     *      a value of <code>null</code> or a (trimmed) length of zero
     */
    protected String stringOrNone(String string) {
        if (string != null && string.trim().length() != 0) {
            return string.trim();
        }
        return "None";
    }
    
    /**
     * The {@link Socials} object.
     * @see {@literal #getPlugin()}
     */
    private final Socials plugin;
}
