/**
 * @file Message.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials.message;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.bukkit.ChatColor;
import org.bukkit.entity.Player;
import org.mudcraft.bukkit.socials.Socials;
import org.mudcraft.bukkit.socials.user.SocialUser;

/**
 * The {@link Message} class is used used to create and broadcast
 * output in accordance with a format that also varies depending upon
 * which player witnesses it.
 * <p>
 * The format specifier is a character string.  The format specifier is
 * composed of zero of more directives: ordinary characters (not {), which
 * are copied, unchanged, to the output; and conversion specifications,
 * each of which results in fetching zero or more arguments.
 * <p>
 * Each conversion specification is introduced by the { character, and
 * ends with the } character.  In between, there may be one or two macros,
 * the second of which, if present, is delimited by a : character.  The
 * first macro is expanded only when the message is witnessed by the
 * actor, the {@link Player} performing the message; the second macro is
 * expanded when another {@link Player} witnesses the message.
 * <p>
 * Macros may contain either one component, or two components separated
 * by a . character.  A single component is interpreted as a literal; it
 * is trimmed of leading and trailing whitespace but no other evaluation
 * is performed.
 * Alternatively, if two components are found, the first identifies the
 * context element to be evaluated and the second identifies which
 * property of the context element to select.
 * <p>
 * The following properties are supported:
 * <ul>
 * <li>gender: the player's gender</li>
 * <li>heSheIt: the objective pronoun for the player's gender</li>
 * <li>himHerIt: the subjective pronoun for the player's gender</li>
 * <li>hisHerIts: the possessive indicator for the player's gender</li>
 * <li>hisHersIts: the possessive pronoun for the player's gender</li>
 * <li>int: treats the context element as an integer</li>
 * <li>intEnglish: same as <i>int</i> but rendered in English</li>
 * <li>name: the player's display name</li>
 * <li>namePossessive: the possessive form of the player's name</li>
 * <li>self: the reflexive pronoun for the player's gender</li>
 * <li>string: treats the context element as a string (trimmed)</li>
 * <li>world: the player's current world</li>
 * </ul>
 * @author Geoffrey Davis
 */
public class Message {
    /**
     * Contains contextual information used to render a message.
     * @author Geoffrey Davis
     */
    public class Context {
        /**
         * Constructs a new {@link Context} instance.
         */
        public Context() {
            // Nothing.
        }
        
        /**
         * Gets whether this object is equal to another object.
         * @param o the object to which this object is to be compared
         * @return <code>true</code> if the specified object is equivalent
         * @see Object#equals(Object)
         */
        public boolean equals(Object o) {
            boolean retcode = false;
            if (o instanceof Context) {
                // Cast the object to a Context instance.
                final Context oContext = (Context) o;
                
                // Compare the two objects for equality.
                retcode = properties.equals(oContext.properties);
            }
            return retcode;
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public double getByte(String key) {
            return getByte(key, (byte) 0);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public double getByte(String key, byte defaultValue) {
            try {
                return Byte.parseByte(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public boolean getBoolean(String key) {
            return getBoolean(key, false);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public boolean getBoolean(String key, boolean defaultValue) {
            final String value = getString(key);
            try {
                return Double.parseDouble(value) != 0.0;
            } catch (NumberFormatException ex) {
                if (value.equals("")) {
                    return defaultValue;
                } else {
                    return Boolean.parseBoolean(value);
                }
            }
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public double getDouble(String key) {
            return getDouble(key, 0.0);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public double getDouble(String key, double defaultValue) {
            try {
                return Double.parseDouble(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public float getFloat(String key) {
            return getFloat(key, 0.0f);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public float getFloat(String key, float defaultValue) {
            try {
                return Float.parseFloat(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public int getInt(String key) {
            return getInt(key, 0);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public int getInt(String key, int defaultValue) {
            try {
                return Integer.parseInt(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public long getLong(String key) {
            return getLong(key, 0);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public long getLong(String key, long defaultValue) {
            try {
                return Long.parseLong(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.
         * @param key the context key
         * @return the {@link Player} whose name equals the value in this
         *      context associated with the specified key, or <code>null</code>
         */
        public Player getPlayer(String key) {
            return getPlugin().getServer().getPlayer(getString(key));
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public short getShort(String key) {
            return getShort(key, (short) 0);
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public short getShort(String key, short defaultValue) {
            try {
                return Short.parseShort(getString(key));
            } catch (NumberFormatException ex) {
                return defaultValue;
            }
        }
        
        /**
         * Searches for the property with the specified key.
         * @param key the context key
         * @return the {@link SocialUser} whose name equals the value in this
         *      context associated with the specified key, or <code>null</code>
         */
        public SocialUser getSocialUser(String key) {
            return getPlugin().loadSocialUser(getString(key));
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public String getString(String key) {
            return getString(key, "");
        }
        
        /**
         * Searches for the property with the specified key.  The method
         * returns the default value if the property is not found.
         * @param key the context key
         * @param defaultValue the default value
         * @return the value in this context associated with the specified
         *      key, or the default value
         */
        public String getString(String key, String defaultValue) {
            return properties.getProperty(key, defaultValue);
        }
        
        /**
         * Calculates a hashcode for this object.
         * @return a hashcode for this object
         * @see Object#hashCode()
         */
        public int hashCode() {
            // Begin with a prime number.
            int hashCode = 433;
            
            // Multiply by a prime number and add `properties` hashcode.
            hashCode = (hashCode * 577) + properties.hashCode();
                
            // Return the calculated hashcode.
            return hashCode;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setBoolean(String key, boolean value) {
            setString(key, Boolean.toString(value));
            return this;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setByte(String key, byte value) {
            setString(key, Byte.toString(value));
            return this;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setDouble(String key, double value) {
            setString(key, Double.toString(value));
            return this;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setFloat(String key, float value) {
            setString(key, Float.toString(value));
            return this;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setInt(String key, int value) {
            setString(key, Integer.toString(value));
            return this;
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setLong(String key, long value) {
            setString(key, Long.toString(value));
            return this;
        }
        
        /**
         * Associates the name of the specified player with the specified key.
         * @param key the context key
         * @param player the player whose name is to be assigned to the
         *      specified key
         * @return a reference to this object
         */
        public Context setPlayer(String key, Player player) {
            return setString(key, player != null ? player.getName() : null);
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setShort(String key, short value) {
            setString(key, Short.toString(value));
            return this;
        }
        
        /**
         * Associates the name of the specified social user with the specified key.
         * @param key the context key
         * @param user the social user whose name is to be assigned to the
         *      specified key
         * @return a reference to this object
         */
        public Context setSocialUser(String key, SocialUser user) {
            return setString(key, user != null ? user.getPlayerName() : null);
        }
        
        /**
         * Associates the specified value with the specified key.
         * @param key the context key
         * @param value the value to be assigned to the specified key
         * @return a reference to this object
         */
        public Context setString(String key, String value) {
            if (key != null) {
                if (value != null) {
                    properties.setProperty(key, value);
                } else {
                    properties.remove(key);
                }
            }
            return this;
        }
        
        /**
         * Contains the properties's key-value pairs.
         */
        private Properties properties = new Properties();
        
        /**
         * A unique version serial number for object serialization.
         */
        private static final long serialVersionUID = 6388673190951775510L;
    }
    
    /**
     * Defines the contract for specialized predicate classes that
     * indicate whether a {@link Player} is eligible to receive a message. 
     * @author Geoffrey Davis
     */
    public static interface Filter {
        /**
         * Returns whether a {@link Player} is accepted by the filter.
         * @param player the {@link Player} object
         * @return <code>true</code> is the specified {@link Player} is
         *      accepted by this filter; <code>false</code> otherwise
         */
        boolean accept(Player player);
    }
    
    /**
     * Constructs a new {@link Message} instance.
     * @param plugin the {@link Socials} object
     */
    public Message(Socials plugin) {
        this.plugin = plugin;
    }
    
    /**
     * Expands a message macro.
     * @param me the {@link Player} who is doing the message
     * @param target the {@link Player} who is witnessing the message
     * @param macro the macro to be expanded
     * @return the macro expansion
     */
    protected String doMacro(Player me, Player target, String macro) {
        // Split the macro on the dot character.
        final String[] dotSplit = macro.split("\\.");
        
        // When we split on dot there are several possible cases
        // that might arise:
        //  1. The string was empty; do nothing.
        //  2. The string contains no dots; we interpret the string as
        //     a literal and return it.
        //  3. The string contains one dot; the first segment is the
        //     object and the second segment is the property.
        //  4. The string contains two or more dots; contextually, this
        //     doesn't mean anything.  We could use the evaluation of
        //     case #3, then recurse, but I'm not certain there is any
        //     value to this.
        
        if (dotSplit.length == 1) {
            // Return the string literal.
            return dotSplit[0];
        } else if (dotSplit.length == 2) {
            // Lookup the player and social user just in case.
            final Player player = context.getPlayer(dotSplit[0]);
            final SocialUser user = context.getSocialUser(dotSplit[0]);
            
            // The `gender` property.
            if (dotSplit[1].equalsIgnoreCase("gender")) {
                if (user != null && user.getGender() != null) {
                    return user.getGender().name().toLowerCase();
                }
            // The `heSheIt` property.
            } else if (dotSplit[1].equalsIgnoreCase("heSheIt")) {
                if (target == me) {
                    return "you";
                } else if (user != null && user.getGender() != null) {
                    return user.getGender().getSubjectivePronoun();
                }
            // The `himHerIt` property.
            } else if (dotSplit[1].equalsIgnoreCase("himHerIt")) {
                if (target == me) {
                    return "you";
                } else if (user != null && user.getGender() != null) {
                    return user.getGender().getObjectivePronoun();
                }
            // The `hisHerIts` property.
            } else if (dotSplit[1].equalsIgnoreCase("hisHerIts")) {
                if (target == me) {
                    return "your";
                } else if (user != null && user.getGender() != null) {
                    return user.getGender().getPossessiveDeterminer();
                }
            // The `hisHersIts` property.
            } else if (dotSplit[1].equalsIgnoreCase("hisHersIts")) {
                if (target == me) {
                    return "yours";
                } else if (user != null && user.getGender() != null) {
                    return user.getGender().getPossessivePronoun();
                }
            // The `int` property.
            } else if (dotSplit[1].equalsIgnoreCase("int")) {
                return Integer.toString(context.getInt(dotSplit[0], 0));
            // The `intEnglish` property (same as `int` right now.)
            } else if (dotSplit[1].equalsIgnoreCase("intEnglish")) {
                return Integer.toString(context.getInt(dotSplit[0], 0));
            // The `name` property.
            } else if (dotSplit[1].equalsIgnoreCase("name")) {
                if (target == me) {
                    return "you";
                } else if (player != null) {
                    if (player.getDisplayName() != null &&
                        player.getDisplayName().trim().length() != 0) {
                        return player.getDisplayName().trim();
                    }
                    return player.getName();
                }
            // The `namePossessive` property.
            } else if (dotSplit[1].equalsIgnoreCase("namePossessive")) {
                if (target == me) {
                    return "your";
                } else if (player != null) {
                    // Assume we'll use the player's name.
                    String displayName = player.getName();
                    
                    if (player.getDisplayName() != null &&
                        player.getDisplayName().trim().length() != 0) {
                        // Use the player's display name instead.
                        displayName = player.getDisplayName().trim();
                    }
                    // Return the possessive form of the name.
                    return displayName +
                        (displayName.endsWith("s") ? "'" : "'s");
                }
            // The `self` property.
            } else if (dotSplit[1].equalsIgnoreCase("self")) {
                if (target == me) {
                    return "yourself";
                } else if (user != null && user.getGender() != null) {
                    return user.getGender().getReflexivePronoun();
                }
            // The `string` property.
            } else if (dotSplit[1].equalsIgnoreCase("string")) {
                return context.getString(dotSplit[0], "").trim();
            // The `world` property.
            } else if (dotSplit[1].equalsIgnoreCase("world")) {
                if (player != null && player.getWorld() != null) {
                    return player.getWorld().getName();
                }
            }
        }
        return "null";
    }
    
    /**
     * Gets the message's chat color.
     * @return the {@link ChatColor} of the message
     * @see #setChatColor(ChatColor)
     */
    public ChatColor getChatColor() {
        return chatColor;
    }
    
    /**
     * Gets the configured filters.
     * @return a mutable {@link List} of {@link Filter}s
     */
    public List<Filter> getFilters() {
        return filters;
    }
    
    /**
     * Gets the message's format specifier.
     * @return the message's format specifier
     * @see #setFormatSpecifier(String)
     */
    public String getFormatSpecifier() {
        return formatSpecifier;
    }
    
    /**
     * Gets the message context.
     * @return the {@link Context} object
     */
    public Context getContext() {
        return context;
    }
    
    /**
     * Gets the Socials plug-in.
     * @return the {@link Socials} object
     */
    public Socials getPlugin() {
        return plugin;
    }
    
    /**
     * Sends the message.
     * @param me the {@link Player} doing the message
     */
    public void send(Player me) {
        // Iterate over the players currently online.
        for (Player player: me.getServer().getOnlinePlayers()) {
            // Indicates whether the player is accepted.
            boolean accepted = true;
            
            // Iterate over the configured filters.
            for (Filter filter: filters) {
                // Determine whether the filter accepts the player.
                accepted = accepted && filter.accept(player);
            }
            if (accepted) {
                // Send the message if the player was accepted.
                send(me, player);
            }
        }
    }
    
    /**
     * Sends the message to a specified player.
     * This method processes the message's format specifier.  Each macro
     * encountered is expanded and the expansion is accumulated into a
     * buffer.  When processing is finished, the message is wrapped and
     * dispatched the the player indicated by <code>target</code>.
     * @param me the {@link Player} who is doing the message
     * @param target the {@link Player} who is witnessing the message
     * @see #send(Player)
     */
    protected void send(Player me, Player target) {
        // Indicates whether we're presenting processing a macro.
        boolean isMacro = false;
        
        // Save off the length of the format specifier.
        final int length = formatSpecifier.length();
        
        // Create some string builders for tracing processing progress.
        final StringBuilder macroBuilder = new StringBuilder();
        final StringBuilder messageBuilder = new StringBuilder(length);
        
        // Iterate over the contents of the format specifier.
        for (int offset = 0; offset < length; ++offset) {
            // Get the character at the current position.
            char ch = formatSpecifier.charAt(offset);
            
            // Check whether we've reached the end of a macro.
            if (isMacro && ch == '}') {
                // Get the macro from the macro builder and trim it.
                final String macro = macroBuilder.toString().trim();
                
                // Split the macro on the colon character.
                final String[] colonSplit = macro.split(":");
                
                // This is the expansion of the message macro.
                String expansion = null;
                
                // Only one term is specified.
                if (colonSplit.length == 1) {
                    // Expand the macro for term #1 (index 0.)
                    expansion = doMacro(me, target, colonSplit[0]);
                } else if (colonSplit.length != 0) {
                    if (target == me) {
                        // Expand the macro for term #1 (index 0.)
                        expansion = doMacro(me, target, colonSplit[0]);
                    } else {
                        // Expand the macro for term #2 (index 1.)
                        expansion = doMacro(me, target, colonSplit[1]);
                    }
                }
                // Make certain the macro expanded to /something/.
                if (expansion != null && expansion.length() != 0) {
                    // Accumulate the macro expansion.
                    messageBuilder.append(expansion);
                    
                    // Determine if we have a configured chat color.
                    if (chatColor != null) {
                        // Accumulate the chat color to prevent bleeding.
                        messageBuilder.append(chatColor);
                    }
                }
                // We're not longer processing a macro.
                isMacro = false;
            // Check if we're accumulating the macro.
            } else if (isMacro) {
                macroBuilder.append(ch);
            // Check if we're starting a macro.
            } else if (ch == '{') {
                // Clear the contents of the macro builder.
                macroBuilder.delete(0, macroBuilder.length());
                
                // Record that we're processing a macro.
                isMacro = true;
            // Otherwise we're accumulating the message.
            } else {
                messageBuilder.append(ch);
            }
        }
        // Was any message actually constructed?
        if (messageBuilder.length() != 0) {
            // Determine if we have a configured chat color.
            if (chatColor != null) {
                // Prepend the message with the chat color.
                messageBuilder.insert(0, chatColor);
            }
            // Use the built-in line wrapper functionality.
            final String[] lines = WordWrap.wrap(messageBuilder.toString());
            
            // Iterate over the wrapped lines.
            for (String line: lines) {
                // Send each line to the target player.
                target.sendMessage(line);
            }
        }
    }
    
    /**
     * Sets the message's chat color.
     * @param chatColor the {@link ChatColor} of the message
     * @return a reference to this object
     * @see #getChatColor()
     */
    public Message setChatColor(ChatColor chatColor) {
        this.chatColor = chatColor != null ? chatColor : ChatColor.WHITE;
        return this;
    }
    
    /**
     * Sets the message's format specifier. 
     * @param formatSpecifier the message's format specifier
     * @return a reference to this object
     * @see #getFormatSpecifier()
     */
    public Message setFormatSpecifier(String formatSpecifier) {
        this.formatSpecifier = null;
        if (formatSpecifier != null && formatSpecifier.trim().length() != 0) {
            this.formatSpecifier = formatSpecifier.trim();
        }
        return this;
    }
    
    protected String[] wrap(String input) {
        return new String[0];
    }
    
    /**
     * The color of the message.
     * @see #getChatColor()
     * @see #setChatColor(ChatColor)
     */
    private ChatColor chatColor = ChatColor.WHITE;
    
    /**
     * The message context.
     * @see #getContext()
     */
    private final Context context = new Context();
    
    /**
     * The configured {@link Filter}s.
     * @see {@link Filter}
     * @see #getFilters()
     */
    private final List<Filter> filters = new ArrayList<Filter>();
    
    /**
     * The message's format specifier.
     * @see #getFormatSpecifier()
     * @see #setFormatSpecifier(String)
     */
    private String formatSpecifier;
    
    /**
     * The {@link Socials} object.
     * @see {@link Message#getPlugin()}
     */
    private final Socials plugin;
}