/**
 * @file WordWrap.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials.message;

import org.bukkit.ChatColor;

/**
 * Wraps text to the width of the client's chat window.
 * @author Geoffrey Davis
 */
public class WordWrap {
    /**
     * Gets the width of a character in pixels. 
     * @param ch the character whose width is to be determined
     * @return the width of the glyph represented by the character
     *      <code>ch</code> in pixels
     */
    private static int widthOf(final char ch) {
        if (ch >= 0 && ch < GLYPH_WIDTH.length) {
            return GLYPH_WIDTH[ch];
        }
        return GLYPH_WIDTH[GLYPH_WIDTH.length - 1];
    }
    
    /**
     * Wraps a string to the width of the client's chat window.
     * This method honors word boundaries.
     * @param input the {@link String} to be wrapped
     * @return an array of {@link String}s, each representing a line of
     *      input no wider than the client's chat window
     */
    public static String[] wrap(final String input) {
        // The last color code encountered.
        char lastColorCode = 'f';
        
        // The width of the current line in pixels.
        int lineWidth = 0;

        // Create a string builder to accumulate each line.
        final StringBuilder stringBuilder = new StringBuilder();
        
        // Iterate over the lines of input.
        for (String line: input.split("\n")) {
            // Indicates whether a sentence is capitalized.
            boolean capitalized = false;
            
            if (lineWidth > 0) {
                // Write a new-line character to the string builder.
                stringBuilder.append('\n');
            }
            // Reset the length of the line.
            lineWidth = 0;
            
            // Iterate over the contents of the input string.
            for (String word: line.split(" ")) {
                // Loop while the word ends with a color indicator character.
                while (word.endsWith("\u00A7")) {
                    // Strip trailing color indicator characters.
                    word = word.substring(0, word.length() - 1);
                }
                
                // Save off the length of the word.
                final int length = word.length();
                
                // The width of the current word in pixels.
                int wordWidth = 0;
                
                // Iterate over the characters of the word.
                for (int i = 0; i < length; ++i) {
                    // Check for the color indicator char.
                    if (word.charAt(i) == '\u00A7') {
                        // Record the color code.
                        lastColorCode = word.charAt(++i);
                    } else {
                        // Calculate the width of the current word.
                        wordWidth += widthOf(word.charAt(i));
                    }
                }
                if (wordWidth != 0) {
                    // Determine whether the word needs to be wrapped.
                    if (lineWidth + wordWidth + widthOf(' ') >= 320) {
                        if (lineWidth > 0) {
                            // Write a new-line character to the string builder.
                            stringBuilder.append('\n');
                        }
                        // Reset the length of the line.
                        lineWidth = 0;
                    }
                    if (lineWidth == 0 && lastColorCode != 'f') {
                        // Write the last color code to the string builder.
                        stringBuilder.append('\u00A7').append(lastColorCode);
                    }
                    if (lineWidth != 0) {
                        // Insert a space before we append the word.
                        stringBuilder.append(' ');
                        
                        // Adjust the width of the line by one space.
                        lineWidth += widthOf(' ');
                    }
                    // Capitalize the word if the sentence is not capitalized.
                    for (int i = 0; !capitalized && i < length; ++i) {
                        // Check whether we found a color.
                        if (word.charAt(i) == '\u00A7') {
                            ++i;
                        } else {
                            // Convert the word to uppercase.
                            word = word.substring(0, i) +
                                   word.substring(i, i + 1).toUpperCase() +
                                   word.substring(i + 1).toLowerCase();
                            
                            // Indicate the sentence has been capitalized.
                            capitalized = true;
                        }
                    }
                    // Append the word to the string builder.
                    stringBuilder.append(word);
    
                    if (capitalized) {
                        // Save off a stripped copy of the word.
                        final String stripped = ChatColor.stripColor(word);
                        
                        // Determine whether the word ends with punctuation.
                        capitalized = stripped.endsWith("!") == false &&
                                      stripped.endsWith(".") == false &&
                                      stripped.endsWith("?") == false;
                    }
                    // Add the width of the word to the width of the line.
                    lineWidth += wordWidth;
                }
            }
        }
        // Convert the contents of the string builder to an array.
        return stringBuilder.toString().split("\n");
    }
    
    /**
     * These are the widths of the character glyphs in pixels.
     */
    private final static int[] GLYPH_WIDTH = new int[] {
        1, 9, 9, 8, 8, 8, 8, 7, 9, 8, 9, 9, 8, 9, 9, 9, 8, 8, 8, 8, 9, 9,
        8, 9, 8, 8, 8, 8, 8, 9, 9, 9, 4, 2, 5, 6, 6, 6, 6, 3, 5, 5, 5, 6,
        2, 6, 2, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 2, 2, 5, 6, 5, 6, 7, 6,
        6, 6, 6, 6, 6, 6, 6, 4, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
        6, 6, 6, 4, 6, 4, 6, 6, 3, 6, 6, 6, 6, 6, 5, 6, 6, 2, 6, 5, 3, 6,
        6, 6, 6, 6, 6, 6, 4, 6, 6, 6, 6, 6, 6, 5, 2, 5, 7, 6, 6, 6, 6, 6,
        6, 6, 6, 6, 6, 6, 6, 4, 6, 3, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
        6, 6, 6, 6, 4, 6, 6, 3, 6, 6, 6, 6, 6, 6, 6, 7, 6, 6, 6, 2, 6, 6,
        8, 9, 9, 6, 6, 6, 8, 8, 6, 8, 8, 8, 8, 8, 6, 6, 9, 9, 9, 9, 9, 9,
        9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 6, 9, 9,
        9, 5, 9, 9, 8, 7, 7, 8, 7, 8, 8, 8, 7, 8, 8, 7, 9, 9, 6, 7, 7, 7,
        7, 7, 9, 6, 7, 8, 7, 6, 6, 9, 7, 6, 7, 1
    };
}
