/**
 * @file SocialUser.java
 * 
 * Copyright (C) 2011 MUDCraft.org
 * All Rights Reserved.
 *
 * @author Geoffrey Davis
 *
 * $Id$
 */
package org.mudcraft.bukkit.socials.user;

import com.avaje.ebean.validation.Length;
import com.avaje.ebean.validation.NotEmpty;

import java.io.Serializable;
import java.util.Date;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

/**
 * Social information for a Minecraft player.
 * @author Geoffrey Davis
 */
@Entity
@Table(name="social_user")
public class SocialUser implements Cloneable, Serializable {
    /**
     * Constructs a new {@link SocialUser} instance.
     */
    public SocialUser() {
        // Nothing.
    }

    /**
     * Constructs a new {@link SocialUser} instance.
     * @param other the {@link SocialUser} whose state is to be copied
     */
    public SocialUser(SocialUser other) {
        if (other != null) {
            this.birthDate        = cloneDate(other.birthDate);
            this.emailAddress     = other.emailAddress;
            this.gender           = other.gender;
            this.id               = other.id;
            this.playerName       = other.playerName;
            this.realGender       = other.realGender;
            this.realName         = other.realName;
            this.showAge          = other.showAge;
            this.showBirthday     = other.showBirthday;
            this.showBirthYear    = other.showBirthYear;
            this.showEmailAddress = other.showEmailAddress;
            this.showRealGender   = other.showRealGender;
            this.showRealName     = other.showRealName;
        }
    }
    
    /**
     * Creates a copy of this object.
     * @return a copy of this object
     * @see Cloneable
     * @see Object#clone()
     */
    public Object clone() {
        return new SocialUser(this);
    }
    
    /**
     * Creates a copy of a {@link Date}.
     * @param date the {@link Date} to be cloned
     * @return a copy of the specified date, or <code>null</code>
     */
    private final static Date cloneDate(Date date) {
        return date != null ? new Date(date.getTime()) : null;
    }
    
    /**
     * Gets whether this object is equal to another object.
     * @param o the object to which this object is to be compared
     * @return <code>true</code> if the specified object is equivalent
     * @see Object#equals(Object)
     */
    public boolean equals(Object o) {
        if (o instanceof SocialUser) {
            // Cast the object to a SocialUser.
            final SocialUser oSocialUser = (SocialUser) o;
            
            if (playerName == oSocialUser.playerName) {
                return true;
            } else if (playerName != null) {
                // Compare our player name to the other SocialUser's.
                if (playerName.equalsIgnoreCase(oSocialUser.playerName))
                    return true;
            } else if (oSocialUser.playerName != null) {
                // Compare the other SocialUser's player name to ours.
                if (oSocialUser.playerName.equalsIgnoreCase(playerName))
                    return true;
            }
        }
        return false;
    }
    
    /**
     * Gets the user's birth date.
     * @return the user's birth date
     * @see #setBirthDate(Date)
     */
    public Date getBirthDate() {
        return cloneDate(birthDate);
    }
    
    /**
     * Gets the user's email address.
     * @return the user's email address
     * @see #setEmailAddress(String)
     */
    public String getEmailAddress() {
        return emailAddress;
    }

    /**
     * Gets the user's gender.
     * @return the user's gender
     * @see #setGender(Gender)
     */
    public Gender getGender() {
        return gender;
    }
    
    /**
     * Gets the record identity.
     * @return the record identity
     * @see #setId(Integer)
     */
    public Integer getId() {
        return id;
    }
    
    /**
     * Gets the user's player name.
     * @return the user's player name
     * @see #setPlayerName(String)
     */
    public String getPlayerName() {
        return playerName;
    }
    
    /**
     * Gets the user's real gender.
     * @return the user's real gender
     * @see #setRealGender(Gender)
     */
    public Gender getRealGender() {
        return realGender;
    }
    
    /**
     * Gets the user's real name.
     * @return the user's real name
     * @see #setRealName(String)
     */
    public String getRealName() {
        return realName;
    }
    

    /**
     * Gets whether the user's age should be shown
     * @return <code>true</code> if the user's age should be shown
     * @see #setShowAge(Boolean)
     */
    public Boolean getShowAge() {
        return showAge;
    }

    /**
     * Gets whether the user's birthday should be shown
     * @return <code>true</code> if the user's birthday should be shown
     * @see #setShowBirthday(Boolean)
     */
    public Boolean getShowBirthday() {
        return showBirthday;
    }

    /**
     * Gets whether the user's birth year should be shown
     * @return <code>true</code> if the user's birth year should be shown
     * @see #setShowBirthYear(Boolean)
     */
    public Boolean getShowBirthYear() {
        return showBirthYear;
    }

    /**
     * Gets whether the user's email address should be shown
     * @return <code>true</code> if the user's email address should be shown
     * @see #setShowEmailAddress(Boolean)
     */
    public Boolean getShowEmailAddress() {
        return showEmailAddress;
    }

    /**
     * Gets whether the user's real gender should be shown
     * @return <code>true</code> if the user's real gender should be shown
     * @see #setShowRealGender(Boolean)
     */
    public Boolean getShowRealGender() {
        return showRealGender;
    }
    
    /**
     * Gets whether the user's real name should be shown
     * @return <code>true</code> if the user's real name should be shown
     * @see #setShowAge(Boolean)
     */
    public Boolean getShowRealName() {
        return showRealName;
    }
    
    /**
     * Calculates a hashcode for this object.
     * @return a hashcode for this object
     * @see Object#hashCode()
     */
    public int hashCode() {
        // Begin with a prime number.
        int hashCode = 193;
        
        // Multiply by a prime number and add `playerName` hashcode.
        hashCode = (hashCode * 967) +
            (playerName != null ? playerName.hashCode() : 0);
            
        // Return the calculated hashcode.
        return hashCode;
    }
        
    /**
     * Sets the user's birth date.
     * @param birthDate the user's birth date
     * @see #getBirthDate()
     */
    public void setBirthDate(Date birthDate) {
        this.birthDate = cloneDate(birthDate);
    }

    /**
     * Sets the user's email address.
     * @param emailAddress the user's email address
     * @see #getEmailAddress()
     */
    public void setEmailAddress(String emailAddress) {
        this.emailAddress = null;
        if (emailAddress != null && emailAddress.trim().length() != 0) {
            this.emailAddress = emailAddress.trim().toLowerCase();
        }
    }
    
    /**
     * Sets the user's gender.
     * @param gender the user's gender 
     * @see #getGender()
     */
    public void setGender(Gender gender) {
        this.gender = gender;
    }
    
    /**
     * Sets the record identity.
     * @param id the record identity
     * @see #getId()
     */
    public void setId(Integer id) {
        this.id = id;
    }
    
    /**
     * Sets the user's player name.
     * @param playerName the user's player name
     * @see #getPlayerName()
     */
    public void setPlayerName(String playerName) {
        this.playerName = null;
        if (playerName != null && playerName.trim().length() != 0) {
            this.playerName = playerName.trim().toLowerCase();
        }
    }
    
    /**
     * Sets the user's real gender.
     * @param realGender the user's real gender 
     * @see #getRealGender()
     */
    public void setRealGender(Gender realGender) {
        this.realGender = realGender;
    }
    
    /**
     * Sets the user's real name.
     * @param realName the user's real name
     * @see #getRealName()
     */
    public void setRealName(String realName) {
        this.realName = null;
        if (realName != null && realName.trim().length() != 0) {
            this.realName = realName.trim();
        }
    }

    /**
     * Sets whether the user's age should be shown.
     * @param showAge indicates whether the user's age should be shown 
     * @see #getShowAge()
     */
    public void setShowAge(Boolean showAge) {
        this.showAge = showAge;
    }

    /**
     * Sets whether the user's birthday should be shown.
     * @param showBirthday indicates whether the user's birthday should
     *      be shown 
     * @see #getShowBirthday()
     */
    public void setShowBirthday(Boolean showBirthday) {
        this.showBirthday = showBirthday;
    }

    /**
     * Sets whether the user's birth year should be shown.
     * @param showBirthYear indicates whether the user's birth year
     *      should be shown 
     * @see #getShowBirthYear()
     */
    public void setShowBirthYear(Boolean showBirthYear) {
        this.showBirthYear = showBirthYear;
    }

    /**
     * Sets whether the user's email address should be shown.
     * @param showEmailAddress indicates whether the user's email
     *      address should be shown 
     * @see #getShowEmailAddress()
     */
    public void setShowEmailAddress(Boolean showEmailAddress) {
        this.showEmailAddress = showEmailAddress;
    }
    
    /**
     * Sets whether the user's real gender should be shown.
     * @param showRealGender indicates whether the user's real gender
     *      should be shown 
     * @see #getShowRealGender()
     */
    public void setShowRealGender(Boolean showRealGender) {
        this.showRealGender = showRealGender;
    }

    /**
     * Sets whether the user's real name should be shown.
     * @param showRealName indicates whether the user's real name should
     *      be shown 
     * @see #getShowRealName()
     */
    public void setShowRealName(Boolean showRealName) {
        this.showRealName = showRealName;
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object
     * @see Object#toString()
     */
    @Override
    public String toString() {
        // Create a string builder. 
        final StringBuilder stringBuilder = new StringBuilder();
        
        // Write the contents of the social user to the string builder.
        stringBuilder.
            append(getClass().getSimpleName()).append('{').
            append("birthDate=").append(birthDate).append(", ").
            append("emailAddress=").append(emailAddress).append(", ").
            append("gender=").append(gender).append(", ").
            append("id=").append(id).append(", ").
            append("playerName=").append(playerName).append(", ").
            append("realGender=").append(realGender).append(", ").
            append("realName=").append(realName).append(", ").
            append("showAge=").append(showAge).append(", ").
            append("showBirthday=").append(showBirthday).append(", ").
            append("showBirthYear=").append(showBirthYear).append(", ").
            append("showEmailAddress=").append(showEmailAddress).append(", ").
            append("showRealGender=").append(showRealGender).append(", ").
            append("showRealName=").append(showRealName).
            append('}');

        // Return the contents of the string builder.
        return stringBuilder.toString();
    }
    
    /**
     * The user's date of birth.
     * @see #getBirthDate()
     * @see #setBirthDate(Date)
     */
    @Temporal(TemporalType.TIMESTAMP)
    private Date birthDate;
    
    /**
     * The user's email address.
     * @see #getEmailAddress()
     * @see #setEmailAddress(String)
     */
    @Length(max=255)
    private String emailAddress;
    
    /**
     * The user's character's gender.
     * @see #getGender()
     * @see #setGender(Gender)
     */
    private Gender gender;
    
    /**
     * The record identity.
     * @see #getId()
     * @see #setId(Integer)
     */
    @Column(unique=true)
    @Id
    private Integer id;
    
    /**
     * The user's player name.
     * @see #getPlayerName()
     * @see #setPlayerName(String)
     */
    @Column(unique=true)
    @Length(max=255)
    @NotEmpty
    private String playerName;
    
    /**
     * The user's real gender.
     * @see #getRealGender()
     * @see #setRealGender(Gender)
     */
    private Gender realGender;
    
    /**
     * The user's real name.
     * @see #getRealName()
     * @see #setRealName(String)
     */
    @Length(max=255)
    private String realName;
    
    /**
     * A unique version serial number for object serialization.
     */
    private static final long serialVersionUID = -7630740623219745646L;

    /**
     * Indicates whether the user's age should be shown.
     * @see #getShowAge()
     * @see #setShowAge(Boolean)
     */
    private Boolean showAge;
    
    /**
     * Indicates whether the user's birthday should be shown.
     * @see #getShowBirthday()
     * @see #setShowBirthday(Boolean)
     */
    private Boolean showBirthday;
    
    /**
     * Indicates whether the user's birth year should be shown.
     * @see #getShowBirthYear()
     * @see #setShowBirthYear(Boolean)
     */
    private Boolean showBirthYear;
    
    /**
     * Indicates whether the user's email address should be shown.
     * @see #getShowEmailAddress()
     * @see #setShowEmailAddress(Boolean)
     */
    private Boolean showEmailAddress;

    /**
     * Indicates whether the user's real gender should be shown.
     * @see #getShowRealGender()
     * @see #setShowRealGender(Boolean)
     */
    private Boolean showRealGender;
    
    /**
     * Indicates whether the user's real name should be shown.
     * @see #getShowRealName()
     * @see #setShowRealName(Boolean)
     */
    private Boolean showRealName;
}
